/*
 * laufschrift.c
 *
 * Ein einfache "Laufschrift" auf dem Ping-Pong Board.
 *
 * Kompilierbar mittels AVR Studio 4 oder WinAVR 
 *
 * Der Sourcecode und das Hexfile drfen frei verwendet werden.
 * Nutzung erfolgt auf eigene Gefahr.
 *
 * Ver.     Date         Author           Comments
 * -------  ----------   --------------   ------------------------------
 * 1.00	    07.11.2009   Sascha Bader     initial 
 * 1.10     03.09.2016   Falk Brunner     clean up & upgrade
 */


#define F_CPU 8000000UL	                       /* CPU Takt (fr delay-Routine) */

/* -----------------------------------------
 * Includes
 * -----------------------------------------*/
#include <avr/io.h>
#include <avr/interrupt.h>                      /* Interruptbehandlungsroutinen (fr Timerinterrupt) */
#include <avr/pgmspace.h>                       /* Hilfsfunktionen um Daten aus dem Flash zu lesen */
#include <avr/sleep.h>                          /* Schlaffunktionen zum Strom sparen*/
#include <util/delay.h>		                    /* Deklaration der Verzgerungsfunktionen (_delay_ms) */
#include "font.h"                               /* Deklaration des Zeichensatzes */

/* -----------------------------------------
 * Defines (Prprozessor Makros)
 * -----------------------------------------*/

#define WIDTH 12                                /* Breite des Displays */
#define HEIGHT 10                               /* Hhe des Displays */
#define FONTWIDTH 8                             /* Breite des Zeichensatzes */
#define FONTHEIGHT 10                           /* Hhe des Zeichensatzes */
#define FRAMERATE 100                           /* Bildwiderholrate /Hz */

#define STORE_PULSE PORTB |= (1 << 2); PORTB &= ~(1 << 2);  // Speicherpuls fr Schieberegister
#define SHIFT_PULSE	PORTB |= (1 << 3); PORTB &= ~(1 << 3);  // Schiebepuls fr Schieberegister
#define SHIFT_DATA_LOW  PORTB &= ~(1 << 4);                 // Eingangsdaten fr Schieberegister
#define SHIFT_DATA_HIGH PORTB |=  (1 << 4);

#define GetPixel(x,y) ( leds[y] &  (1<<x))      /* Makro: Ein "Pixel" auslesen */
#define SetPixel(x,y)   leds[y] |=  1<<x        /* Makro: Ein "Pixel" setzen */
#define ClearPixel(x,y) leds[y] &=~(1<<x)       /* Makro: Ein "Pixel" lschen */

/* -----------------------------------------
 * Globale Variablen
 * -----------------------------------------*/
volatile uint16_t leds[WIDTH];                  /* Puffer der LED-Matrix */
volatile uint8_t flag_frame;                    /* Sychronsignal zwischen Multiplexing und Laufschrift */ 
volatile int8_t scroll_period = 20;             /* Periodendauer fr einen Schritt der Laufschrift */
char * tpos;

/* -----------------------------------------
 * Text der Laufschrift
 * -----------------------------------------*/
const char text[] PROGMEM =
" Es gibt eine Theorie, die besagt, wenn jemals irgendwer genau herausfindet,"
" wozu das Universum da ist und warum es da ist, dann verschwindet es auf der Stelle"
" und wird durch noch etwas Bizarreres und Unbegreiflicheres ersetzt."
" - Es gibt eine andere Theorie, nach der das schon passiert ist.   " 
" Und jetzt die Sonderzeichen 0x01-0x1F "
     " \x01 \x02 \x03 \x04 \x05 \x06 \x07 \x08 \x09 \x0A \x0B \x0C \x0D \x0E \x0F"
" \x10 \x11 \x12 \x13 \x14 \x15 \x16 \x17 \x18 \x19 \x1A \x1B \x1C \x1D \x1E \x1F";


/* -----------------------------------------
 * Prototypen der Funktionen
 * -----------------------------------------*/
uint16_t getNextColumn(char c, int pixel_x, int shift_y);
void scrollLeft(void);
void scrollUp(char c, int pixel_y);
void init_adc(void);
uint8_t read_adc(uint8_t channel);
void laufschrift_horizontal(void);
void laufschrift_vertikal(void);

/* -------------------------------------------------------------------------
 * Main Funktion
 *
 * Initialisiert den Timer Interrupt und
 * behandelt die Laufschrift
 * -------------------------------------------------------------------------*/
int main(void) {

	/*---------------------------------------------------
	 * Ports konfigurieren (Ein-/Ausgnge)
	 *---------------------------------------------------*/

	DDRC = 0x0f;   // ( PORTC4-7 als ADC-Eingang)
	DDRB = 0xff;   //  PortB0-7 = Output
	DDRD = 0xff;   //  PortD0-7 = Output

    init_adc();

	/*---------------------------------------------------------------------------
	 * 16-Bit Timer 1 fr das Multiplexing der LEDs initialisieren
	 *---------------------------------------------------------------------------*/

	TCCR1B |= (1<<WGM12) | (1<<CS11);   // mode 4, CTC, Vorteiler /8
	TIMSK  |= (1<<OCIE1A); 				// OCR1A Interrupt
    OCR1A   = F_CPU / (8 * 12 * FRAMERATE) -1 ;

	sei();							    // Interrupts global erlauben

	/*---------------------------------------------------
	 * Hauptschleife (Laufschrift erzeugen)
	 *---------------------------------------------------*/
    tpos=(char*)text;

    while(1) {
        if (read_adc(7) > 128) {                // rechtes Poti abfragen, Auswahl der Laufschrift
            laufschrift_horizontal();
        } else {
            laufschrift_vertikal();
        }
    }
}

/* -------------------------------------------------------------------------
 * Funktion getNextColumn
 *
 * Berechnet einen 1 Pixel breitem Ausschnitt des Lauftextes (rechte Spalte),
 * welche rechts hereinscrollt
 * 
 * \param c       Auszugebendes Zeichen
 * \param pixel_x Auszugebende Spalte des Zeichens
 * \param shift_y Vertikale Verschiebung des Ausgabezeichens
 * -------------------------------------------------------------------------*/
uint16_t getNextColumn(char c, int pixel_x, int shift_y) {
    char fontbyte;
    char* tmp_font = (char*)font + (int)c * FONTHEIGHT;
    uint8_t y;
    uint8_t in_mask;            // Bitmaske zum Prfen der Fontdaten
    uint16_t out_mask;          // Bitmaske zum Setzen der Ausgangsdaten
    uint16_t spalte=0;          // neue Spalte

    out_mask = 1<<shift_y;             
    in_mask  = 1<<pixel_x;

    for (y=0; y<FONTHEIGHT; y++) {
	    fontbyte = pgm_read_byte(tmp_font++);       /* Ein Byte (Zeile) des aktuellen Zeichens lesen */
        if (fontbyte & in_mask) {                   /* Prfen ob das Bit in der Spalte des Zeichens gesetzt ist */
            spalte |= out_mask;                     /* Setzen eines Pixels in der Ausgangsvariable */
        }
        out_mask<<=1;
    }

    return spalte;
}

/* -------------------------------------------------------------------------
 * Funktion ScrollLeft
 *
 * Verschiebt den Inhalt der LED-Matrix um eine Spalte nach links.
 * Die letzten Spalte bleibt erhalten
 * -------------------------------------------------------------------------*/
void scrollLeft(void) {
  uint8_t x;                           /* Spaltenzhler */

  for (x=0; x<WIDTH-1; x++) {
	  leds[x]=leds[x+1];                /* Spalten nach links verschieben */
  }
}

/* -------------------------------------------------------------------------
 * Funktion ScrollUp
 *
 * Verschiebt den Inhalt der LED-Matrix um eine Zeile nach oben.
 * und fgt unten eine neue an
 * -------------------------------------------------------------------------*/
void scrollUp(char c, int pixel_y) {
    uint8_t x;                           /* Spaltenzhler */
    char* tmp_font = (char*)font + (int)c * FONTHEIGHT + pixel_y;    
    uint8_t char_line;
    uint8_t in_mask = 0x1;

    char_line = pgm_read_byte(tmp_font);

    leds[0] = leds[0]>>1;
    leds[1] = leds[1]>>1;
    leds[10] = leds[10]>>1;
    leds[11] = leds[11]>>1;

    for (x=2; x<10; in_mask<<=1, x++) {     // FONT-Daten in Spalten 2-9 einfgen
        if (char_line & in_mask) {
	        leds[x] = leds[x]>>1 | 0x200; 
        } else { 
            leds[x] = leds[x]>>1;
        }
    }
}

void init_adc(void) {
    ADMUX  = (1<<REFS0) | (1<<ADLAR);   // ref=AVCC, Ergebnis links ausgerichtet
    ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1);   //  ADC ein, Vorteiler 64
    read_adc(0);                        // erste Messung ist fehlerhaft
}

uint8_t read_adc(uint8_t channel){
    channel &= 0x7;                     // Kanal maskieren, nur 0-7 gltig
    ADMUX = (ADMUX & ~0x7) | channel;   // Kanal auswhlen
    ADCSRA |= (1<<ADSC);                // Messung starten
    while( ADCSRA & (1<<ADSC) );        // warte auf das Ende der Wandlung
    return ADCH;                        // Ergebnis zurckgeben (8 Bit)
}

void laufschrift_horizontal(void) {
	uint8_t x;
    uint16_t neue_spalte;

    while(1) {                                                      // Endlosschleife
		for (; pgm_read_byte(tpos)!='\0'; tpos++) { // Aktuelles Zeichen lesen
	        for (x=0; x < FONTWIDTH; x++) {                         // Pixel des Zeichens abarbeiten
                neue_spalte = getNextColumn(pgm_read_byte(tpos), x, 0); // Ganz rechts eine Spalte des Zeichens berechnen
                while(!flag_frame) {                                // warte auf Sycnchronisation
                    set_sleep_mode(SLEEP_MODE_IDLE);                // schlafen spart Strom
                    sleep_mode();
                }                             
                flag_frame =0;
				scrollLeft();                                       // Platz schaffen und Zeilen nach links schieben
				leds[WIDTH-1] = neue_spalte;                        // neue Daten in Anzeigepuffer speichern       
                scroll_period = (read_adc(6)>>3)+1;                 // Laufgeschwindigkeit per linkem Poti einstellen
                if (read_adc(7) < 128) return;                      // anderer Modus, Funktion verlassen
	        }
		}
        tpos=(char*)text;
    }
}

void laufschrift_vertikal(void) {

	int8_t y;

    while(1) {                                                      // Endlosschleife
		for (; pgm_read_byte(tpos)!='\0'; tpos++) {                 // Aktuelles Zeichen lesen
	        for (y=0; y < FONTHEIGHT; y++) {                        // Zeichens zeilenweise abarbeiten
                while(!flag_frame) {                                // warte auf Sycnchronisation
                    set_sleep_mode(SLEEP_MODE_IDLE);                // schlafen spart Strom
                    sleep_mode();
                }                             
                flag_frame =0;
				scrollUp(pgm_read_byte(tpos), y);                   // nach oben scrollen
                scroll_period = (read_adc(6)>>3)+1;                 // Laufgeschwindigkeit per linkem Poti einstellen
                if (read_adc(7) > 128) return;                      // anderer Modus, Funktion verlassen
            }
		}
        tpos=(char*)text;
    }
}

/* -------------------------------------------------------------------------
 * Interrupt Routine, Multiplexing der LED-Matrix
 *
 * Gibt nacheinander alle Spalten mit LED-Daten aus.
 * Dazu wird mittels der Schieberegister die aktuelle Spalte
 * ausgewhlt und dann das Bitmuster derselben auf die Ports
 * gegeben. Beim nchsten Interrupt ist dann die nchste Spalte dran.
 * -------------------------------------------------------------------------*/

ISR(TIMER1_COMPA_vect) {
	uint16_t leddata;
    static uint8_t spalte;
    static uint8_t frame;

	/*--------------------------------------------------
	 * neue Spalte berechenen
	 *--------------------------------------------------*/

   	spalte++;
	if (spalte == WIDTH) {
		spalte = 0;
        SHIFT_DATA_LOW
	} else {
        SHIFT_DATA_HIGH
    }

	/*------------------------------------------------------------------------
	 * Synchronsignal fr Laufschrift generieren
     * Dadurch kommt es nicht zur ungewollten berschneidungen von
     * Multiplexing und Update der LED-Daten, was zu Ruckeln und Flackern in
     * der Anzeige fuhren kann. Auerdem ist damit die Schrollgeschwindigkeit
     * konstanter und genauer, denn _delay_ms() wird durch die Multiplex-ISR
     * immer wieder unterbrochen, sodass die Zeit nicht exakt dem Parameter
     * entspricht.   
     * Nach dem Ende der ISR und dem Anfang der nchsten ISR kann in der
     * Hauptschleife die LED-Matrix mit neuen Daten versehen werden, ohne da
     * es zu Anzeigefehlern kommt.
	 *----------------------------------------------------------------------*/

    if (spalte == WIDTH-1) {
        frame++;
        if (frame == scroll_period) {
            frame = 0;
            flag_frame = 1;
        }
    }

	/*--------------------------------------------------
	 * altes Spaltenmuster lschen
	 *--------------------------------------------------*/

	PORTC &= ~0x0F;
	PORTD &= ~0xF0;
	PORTB &= ~0x03;

	/*---------------------------------------------------
	 * Daten der neuen Spalte holen
	 *---------------------------------------------------*/

	leddata = leds[spalte];

	/*---------------------------------------------------
	 * Impulse fr die Schieberegister generieren
     * neue Spalte einschalten
	 *---------------------------------------------------*/

    SHIFT_PULSE
    STORE_PULSE

	/*---------------------------------------------------
	 * neue Spaltendaten ausgeben
	 *---------------------------------------------------*/

	PORTC |= leddata & 0x0F;
	PORTD |= leddata & 0xF0;
	PORTB |= (leddata >> 8) & 0x03;
}
