/*

Uhr mit graphischer Anzeige auf PONG-Hardware

*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>

#define F_CPU 8000000                   // CPU Takt in Hz
#define FRAMERATE 100                   // Bildwiederholrate fr LED-Matrix

#define DATA_SET    PORTB |=  (1<<4)    // serielle Daten fuer Schieberegister
#define DATA_CLR    PORTB &= ~(1<<4)    // 
#define CLK_SET     PORTB |=  (1<<3)    // Takt fuer Schieberegister
#define CLK_CLR     PORTB &= ~(1<<3)
#define STROBE_SET  PORTB |=  (1<<2)    // speichern der seriellen Daten im Ausgangsregister
#define STROBE_CLR  PORTB &= ~(1<<2)

// Kodierung der LEDs fr die Ziffern 0-9,
// Matrixanzeige 3x5,
// Ansicht 90 Grad nach rechts gedreht

const char hex2led[10][3] PROGMEM = {

{  // 0
    0b11111,
    0b10001,
    0b11111
},

{  // 1
    0b00100,
    0b00010,
    0b11111
},

{  // 2
    0b11101,
    0b10101,
    0b10111
},

{  // 3
    0b10101,
    0b10101,
    0b11111
},

{  // 4
    0b00111,
    0b00100,
    0b11111
},

{  // 5
    0b10111,
    0b10101,
    0b11101
},

{  // 6
    0b11111,
    0b10100,
    0b11100
},

{  // 7
    0b00001,
    0b00001,
    0b11111
},

{  // 8
    0b11111,
    0b10101,
    0b11111
},

{  // 9
    0b00111,
    0b00101,
    0b11111
}

};

#define SekundeOn  0b0101000000
#define SekundeOff 0

// Variablen

volatile uint16_t led[12];

volatile uint8_t stunden  = 13;
volatile uint8_t minuten  = 50;
volatile uint8_t sekunden = 33;
volatile uint8_t flag_1s;

// zeichnet eine Ziffer in die LED-Matrix

void zeichne_ziffer(int x, int y, PGM_P ziffer) {
	uint8_t i;
	uint16_t mask, new_data;

	mask = ~(0b11111 << y);

	// neue Ziffer reinkopieren, dabei die alten LEDs loeschen
	for (i=0; i<3; i++) {
		new_data = pgm_read_byte(&ziffer[i]);
		led[i+x] = (led[i+x] & mask) | (new_data<<y);
	}

}

void Anzeige(void)
{ 
    zeichne_ziffer(2,0, hex2led[stunden/10]);
    zeichne_ziffer(6,0, hex2led[stunden%10]);
    zeichne_ziffer(2,5, hex2led[minuten/10]);
    zeichne_ziffer(6,5, hex2led[minuten%10]);

    if (sekunden & 1) {
        led[5] = SekundeOn;
    } else {
        led[5] = SekundeOff;
    }

/*
   zeichne_ziffer(0,0, hex2led[stunden/10]);
   zeichne_ziffer(3,0, hex2led[stunden%10]);
   zeichne_ziffer(6,0, hex2led[minuten/10]);
   zeichne_ziffer(9,0, hex2led[minuten%10]);
   zeichne_ziffer(3,5, hex2led[sekunden/10]);
   zeichne_ziffer(7,5, hex2led[sekunden%10]);
*/
}

void inittimer()
{
    // Timer 2 ,32768 Hz, 8Bit, Prescaler 128 -> 1s
/*    
    ASSR|=(1<<AS2);                // Asynchronmodus
    TCCR2=(1<<CS22) | (1<<CS20);   // Prescaler 128
    TIMSK|=(1<<TOIE2);             // enable Interrupt
*/
    // Timer1, mode 4, CTC, Prescaler=1 
    TCCR1B = (1<<WGM12) | (1<<CS10);
    TIMSK  |= (1<<OCIE1A);              // OCR1A Interrupt enable
    OCR1A = (F_CPU / (12 * FRAMERATE))-1;
}

// Timer 2, 1s

ISR(TIMER2_OVF_vect)
{
    flag_1s=1;
    sekunden++;
    if(sekunden==60)
    {
        sekunden=0;
        minuten++;
        if(minuten==60)
        {
            minuten=0;
            stunden++;
            if(stunden==24)
            {
                stunden=0;
            }
        }
    }
}

// Timer 1, 830us fr LED-Multiplexing

ISR(TIMER1_COMPA_vect)
{
    uint16_t spalte=0;
    static uint8_t count=0;
    static uint8_t count1s=0;

    // Spaltenmultiplexing ber Schieberegister
    count++;
    if (count==12)
    {
        count=0;
        DATA_CLR;
        count1s++;
        if (count1s==FRAMERATE) {
            count1s = 0;
            flag_1s = 1;
        }
    }
    else
    {
        DATA_SET;
    }
    
    // altes Spaltenmuster lschen
    PORTB &= ~0x03;
    PORTC &= ~0x0F; 
    PORTD &= ~0xF0;

    // neues Muster lesen
    spalte=led[count];

    // Spalten-Mux einen Takt weiter schieben
    CLK_SET;
    CLK_CLR;

    // neue Spalte einschalten           
    STROBE_SET;
    STROBE_CLR;
    
    // neues Spaltenmuster ausgeben
    PORTB |= (spalte >> 8) & 0x03;
    PORTC |= spalte & 0x0F; 
    PORTD |= spalte & 0xF0;

} 

int main(void)
{ 
    DDRC = 0x0f; // PC0-PC3 (R1-R9) zu Ausgnge
    DDRB = 0xff; // PD4-PD7 (R5-R8) zu Ausgnge
    DDRD = 0xff; // PB0-PB1 (R9-R10() zu Ausgnge

    //ACSR = 0x80;

    PORTD = 0;   // alle Bits auf 0
    PORTB = 0; 
    PORTC = 0;

    inittimer();

    sei();

    while(1)
    {
		set_sleep_mode(SLEEP_MODE_IDLE);
		sleep_mode();

        if (flag_1s) {
            flag_1s = 0;
            sekunden++;

            if(sekunden==60)
            {
                sekunden=0;
                minuten++;
                if(minuten==60)
                {
                    minuten=0;
                    stunden++;
                    if(stunden==24)
                    {
                        stunden=0;
                    }
                }
            }
            Anzeige();
        }
    }
}

